#include <intercomm.h>
#include <pvm3.h>
#include <stdio.h>
#include <stdlib.h>
 
IC_Desc* create_ttable(int p, int r, float A[200]) {
  int globals[200], offsets[200], tasks[200];
  IC_Desc* desc;
  int i;
  
  /* assign this processor the rth portion of global index space */
  for (i = 0; i < 200; ++i) {
    globals[i] = 200*r + i;
  }
  
  /* round-robin global index to task assignment */
  for (i = 0; i < 200; ++i) {
    offsets[i] = i/p + 50*r;
    tasks[i] = i%p;
  }
  
  /* create the ic descriptor */
  desc = IC_Create_ttable_desc(globals, offsets, tasks, 200);
  
  /* assign each local value its global index */
  for (i = 0; i < 200; ++i) {
    A[i] = globals[i];
  }
  
  return desc;
}


/* copied and modified from jylee's pvm_start.c code
 */
int start_master_slave(char **arg) {
  int numNode, i, j;
  int tids[100];
  int erm;
  int parent;

  FILE* fp;
  char nodename[100]; // restriction: the max. nodename to 100 characters

  if (pvm_parent() < 0) {
    fprintf(stdout,"Spawning %s on %s hosts\n", arg[0], arg[3]);
    sscanf(arg[3], "%i", &numNode);

    if (numNode == 1)
      return numNode;

    if( (fp = fopen(arg[5], "r")) == NULL){
      fprintf(stdout, "machine file open error\n");
      return 0;
    }


    if(fgets(nodename, 100, fp) == NULL){
      return 0;
    }
      

    i = 0;
    while(fgets(nodename, 100, fp) != NULL){
      nodename[strlen(nodename)-1] = 0; // cut newline character 
      fprintf(stderr, "name: <%s>\n", nodename);

      erm = pvm_spawn(arg[0], arg+1, PvmTaskHost, nodename, 1, &tids[i]);
      i++;

      if(erm != 1){
	pvm_perror("Error occurred in spawning...\n");

	for (j=2 ; j<numNode-1 ; j++){
	  if(tids[j] > 0){
	    pvm_kill(tids[j]);
	  }
	}
	pvm_exit();
	return 0;
      }
      
    }


    fclose(fp);

    pvm_initsend(PvmDataDefault);
    pvm_pkint(&numNode, 1, 1);
    if (erm = pvm_mcast(tids, numNode-1, 001) < 0){
      pvm_perror("error in first mcast\n");
    }
    printf("\nStarted %i tasks (one for each node)\n", numNode);


    /*
     * Code to take numNode-1 processes at once.
     *
    erm = pvm_spawn(arg[0], arg+1, PvmTaskDefault, NULL, numNode-1, tids);
    printf("Started %i task(s), total task(s) %i\n", erm, numNode);
    if (erm < numNode-1) {
      pvm_perror("spawn\n");
      for (i = 0; i < numNode-1; i++)
        if (tids[i] > 0)
          pvm_kill(tids[i]);
      pvm_exit();
      return 0;
    }
    else {
      pvm_initsend(PvmDataDefault);
      pvm_pkint(&numNode, 1, 1);
      if (erm = pvm_mcast(tids, numNode-1, 001) < 0)
        pvm_perror("error in first mcast\n");
      printf(" Started %i task\n", erm);
    }
    */
  }
  else {
    parent = pvm_parent();
    printf("parent detected\n");
    pvm_recv(parent, 001);
    pvm_upkint(&numNode, 1, 1);
    sleep(2);
  }

  return numNode;
}


/*
 * parameter
 *  argv[0] : program name
 *  argv[1] : local name
 *  argv[2] : other name
 *  argv[3] : local tasks
 *  argv[4] : other tasks
 *  argv[5] : machine file
 */
int main(int argc, char* argv[]) {
  char* localname; // = argv[1];
  char* othername; // = argv[2];
  int local_tasks; // = atoi(argv[3]);
  int other_tasks; // = atoi(argv[4]);
  int i, rank;
  
  char* groupcomm = "local_c";

  IC_Program* local;
  IC_Program* other;
  
  IC_Desc* desc;
  float A[200];
  
  IC_Region* region_set[2];
  int indices[8];
  IC_Sched* sched;
  int tag = 99;

  int numNode = -1;
  char** newargv;

  
  /* initialize pvm */
  printf("initialize pvm\n");
  pvm_catchout(stdout);

  localname   = argv[1];
  othername   = argv[2];
  local_tasks = atoi(argv[3]);
  other_tasks = atoi(argv[4]);

  numNode = start_master_slave(argv);
  printf("%d nodes \n", numNode);
  rank = pvm_joingroup(groupcomm); 
  pvm_barrier(groupcomm, local_tasks);

  
  /* initialize ic */
  printf("initialize ic\n");
  local = IC_Init(localname, local_tasks, rank);
  other = IC_Wait(othername, other_tasks);
  IC_Sync(local, other);

  desc = create_ttable(local_tasks, rank, A);
 
  /* define two 8 element regions for transfer */
  printf("define regions\n");
  for (i = 0; i < 8; ++i)
    indices[i] = i;
  region_set[0] = IC_Create_enum_region(indices, 8);
  for (i = 0; i < 8; ++i)
    indices[i] = i + 400;
  region_set[1] = IC_Create_enum_region(indices, 8);
  
  /* create a translation table type array descriptor and array */
  printf("create schedule\n");
  sched = IC_Compute_schedule(local, other, desc, region_set, 2);
  printf("send data\n");
  IC_Send_float(other, sched, &A[0], tag);
  IC_Sync(local, other);
  IC_Free_sched(sched);
  
  /* clean up */
  printf("clean up\n");
  IC_Free_region(region_set[1]);
  IC_Free_region(region_set[0]);

  IC_Free_desc(desc);

  IC_Free_program(other);
  IC_Quit(local);

  pvm_barrier(groupcomm, local_tasks);
  pvm_lvgroup(groupcomm);
  pvm_exit();
  
  return 0;
}
