/**
 * $Author: normanlo $
 * $Date: 2006/05/04 16:36:35 $
 * $Revision: 1.1.2.8 $
 * $Source: /fs/schisto7/parti/repository/InterComm/examples/rainbow/Attic/violet.c,v $
 * 
 * Written by Il-Chul Yoon <iyoon@cs.umd.edu>
 */

#include <intercomm.h>
#include <stdio.h>
#include <unistd.h>
#include "pvm_tls.h"

#define PVM_GROUP      "VioletPvm"
#define COMPONENT_NAME "violetcomp"
#define REGION_NAME1   "violetR1"
#define REGION_NAME2   "violetR2"
#define REGION_NAME3   "violetR3"

/**
 * Local view
 * Initialize 2-dim float array
 */
void init_data(int rank, float data[2][2]) {
    int i,j;

    fprintf(stderr, "initialize data\n");
    for (i = 0 ;  i < 2 ; i++){
        for (j = 0 ; j < 2 ; j++){
            data[i][j] = rank*100 + i*10 + j;
        }
    }
}

void print_array(int rank, float data[2][2]){
    int i, j;
    
    fprintf(stdout, "My rank is : %d\n", rank);
    for( i = 0 ; i < 2 ; i++ ){
        for( j = 0 ; j < 2 ; j++){
            fprintf(stdout, "%4.0f  ", data[i][j]);
        }
        fprintf(stdout, "\n");
    }
    fprintf(stdout, "\n");
}

int parse_xjdname(int argc, char** argv, char** xjdname){
    int c; 
    int xjdflag = 0;

    optind = 1; /* reset option index to 1 to parse from the first option*/
   
    while ((c = getopt (argc, argv, "m:n:d:")) != -1){
        switch(c){
            case 'd':
                *xjdname = (char*)strdup(optarg);

                xjdflag = 1;
                break;
        }
    }
    
    if(!xjdflag){
        return -1;
    }
    else{
        return 0; 
    }
}


/**
 * Global view
 * This part should be decided by application programmer
 */
IC_Desc *create_block() {
    int blocks[8][2][2] = {
        { {0, 0}, {1, 1} },
        { {0, 2}, {1, 3} },
        { {2, 0}, {3, 1} },
        { {2, 2}, {3, 3} },
        { {4, 0}, {5, 1} },
        { {4, 2}, {5, 3} },
        { {6, 0}, {7, 1} },
        { {6, 2}, {7, 3} }
    };
    
    int tasks[8]= {0, 1, 2, 3, 4, 5, 6, 7};
    
    return IC_Create_bdecomp_desc(2, &blocks[0][0][0], tasks, 8, IC_ROW_MAJOR);
}

/**
 * This part should be decided by application programmer
 */

void create_region2(IC_Region* second_region[2]){
    int lower[2], upper[2], stride[2];
    
    lower[0] = 2;
    lower[1] = 0;
    upper[0] = 3;
    upper[1] = 1;
    stride[0] = 1;
    stride[1] = 1;
    second_region[0] = IC_Create_block_region(2, lower, upper, stride);

    lower[0] = 6;
    lower[1] = 2;
    upper[0] = 7;
    upper[1] = 2;
    stride[0] = 1;
    stride[1] = 1;
    second_region[1] = IC_Create_block_region(2, lower, upper, stride);
}

void finishPVM(char* grpname, int my_nTask){
    pvm_barrier(grpname, my_nTask);
    pvm_lvgroup(grpname);
    pvm_exit();
}

/*
 * execution
 * $PVM_BIN/violet -n 8 -d full_path_name_of_xjd.xml 
 * 
 * this component expose...
 *  name: violetcomp
 *  dataport: violetR1
 */
int main(int argc, char **argv) {
    
    IC_XJD* xjd;
    
    float local_data[2][2];
    IC_Desc *desc;

    IC_Region* second_region[2];

    int bcast_recvtest[1];
    int bcasttest[10], i;

    int my_nTask, rank, status;
    char* xjdname;
  
    if(parse_xjdname(argc, argv, &xjdname) < 0){
        fprintf(stderr, "XJD Option is required...\n");
	return -1;
    }

    fprintf(stderr, "initialize pvm and getting rank\n");
    if( (init_pvm_mfile(argc, argv, PVM_GROUP, &my_nTask, &rank) ) < 0){
      fprintf(stderr, "Error in initializing PVM...\n");
      return -1;
    }


    fprintf(stderr, "initialize InterComm\n");    
    if ( (xjd = IC_Initialize(COMPONENT_NAME, rank, xjdname, &status)) == NULL ){
        fprintf(stderr, "IC_Initialize failed at rank : %d\n", rank);        
        return -1;
    }
    
    /* Initialize local block and set global view*/
    init_data(rank, local_data);
    desc = create_block();

    /* register regions    */
    create_region2(second_region);
    IC_Register_region(xjd, second_region, 2, REGION_NAME2, desc, local_data, &status);

    /* all regions are registered. Now.. we can export or import    */
    IC_Commit_region(xjd, &status);

    /* receive regions*/
    status = IC_Import(xjd, REGION_NAME2);

    /* print result array */
    print_array(rank, local_data);

    /* receive the broadcasted 1-cell array*/
    status = IC_Recv_Local(xjd, REGION_NAME1, bcast_recvtest, 1);
    fprintf(stderr, "Received Bcast Value: %d\n", bcast_recvtest[0]);
    

    /* broadcast 10-cell array   */
    for(i=0; i<10 ; i++){
        bcasttest[i] = i;
    }
    status = IC_BCast_Local(xjd, REGION_NAME3, bcasttest, 10);

 
    /* free allocated memory*/
    fprintf(stderr, "clean up\n");
    IC_Free_region(second_region[0]);
    IC_Free_region(second_region[1]);
    IC_Free_desc(desc);
    IC_Finalize(xjd, &status);
    finishPVM(PVM_GROUP, my_nTask);
    
    return 0;
}
