divert(-1)
/*//// M4 macros for P++ methods*/

define(`PPP_EXPORT_HELPER',`
void IC_EndPoint::exportArray(const $1Array& a, int msgtag, int& status) {
	#define IC_DEBUG
/* exportArrays a double array using InterComm*/
  unsigned ndim=a.numberOfDimensions();
  if (ndim>IC_MAX_DIM) {
#ifdef IC_DEBUG
    cout << IC_EndPoint::IC_errors[-IC_INVALID_NDIM] << endl;
#endif
    status=IC_INVALID_NDIM;
    return;
  }
  unsigned i;
  int lower[IC_MAX_DIM], upper[IC_MAX_DIM], stride[IC_MAX_DIM];
  
  for(i=0;i<ndim;++i) {
    lower[i]=a.getBase(i); 
    upper[i]=a.getBound(i); 
    stride[i]=a.getStride(i);

#ifdef DEBUG
    cout << "P++ArrayInfo : " << lower[i] << " " << upper[i] << " " << stride[i] << endl;
#endif
  }

#ifdef DEBUG
  cout << "--- number of dimensions " << ndim << endl;
#endif


  IC_Region* region_set[1];
  region_set[0]=IC_Create_block_region(ndim,lower,upper,stride);
  if (!region_set[0]) {
#ifdef IC_DEBUG
    cout << IC_EndPoint::IC_errors[-IC_CANT_ALLOC_REGION] << endl;
#endif
    status=IC_CANT_ALLOC_REGION;
    return;
  }

  IC_Desc* desc=getPartiDescriptor(a); /*,thisProgramAO==otherProgramAO);*/
  if (!desc) {
#ifdef IC_DEBUG
    cout << IC_EndPoint::IC_errors[-IC_CANT_GET_DA_DESCRIPTOR] << endl;
#endif
    status=IC_CANT_GET_DA_DESCRIPTOR;
    return;
  }

  /* this can probably be cached in a more sophisticated version!!!*/
  IC_Sched* sched=IC_Compute_schedule(thisProgram,otherProgram,desc,region_set,1);
  if (!sched) {
#ifdef IC_DEBUG
    cout << IC_EndPoint::IC_errors[-IC_CANT_COMPUTE_COMM_SCHEDULE] << endl;
#endif
    status=IC_CANT_COMPUTE_COMM_SCHEDULE;
    return;
  }

  int ret=IC_Send_$1(otherProgram,sched,a.getLocalArray().getDataPointer(), msgtag);
  if (ret!=0) {
#ifdef IC_DEBUG
    cout << IC_EndPoint::IC_errors[-IC_COMM_FAILURE] << endl;
#endif
    status=IC_COMM_FAILURE;
    return;
  }

  IC_Free_region(region_set[0]);
  IC_Free_desc(desc);

  status=IC_OK;
}')

define(`PPP_IMPORT_HELPER',`
void IC_EndPoint::importArray(const $1Array& a, int msgtag, int& status) {
/* importArrays a double array using InterComm*/
  unsigned ndim=a.numberOfDimensions();
  if (ndim>IC_MAX_DIM) {
#ifdef IC_DEBUG
    cout << IC_EndPoint::IC_errors[-IC_INVALID_NDIM] << endl;
#endif
    status=IC_INVALID_NDIM;
    return;
  }
  unsigned i;
  int lower[IC_MAX_DIM], upper[IC_MAX_DIM], stride[IC_MAX_DIM];

  for(i=0;i<ndim;++i) {
    lower[i]=a.getBase(i); 
    upper[i]=a.getBound(i); 
    stride[i]=a.getStride(i);

#ifdef DEBUG
    cout << "P++ArrayInfo : " << lower[i] << " " << upper[i] << " " << stride[i] << endl;
#endif
  }

#ifdef DEBUG
  cout << "--- number of dimensions is " << ndim << endl;
#endif

  IC_Region* region_set[1];
  region_set[0]=IC_Create_block_region(ndim,lower,upper,stride);
  if (!region_set[0]) {
#ifdef IC_DEBUG
    cout << IC_EndPoint::IC_errors[-IC_CANT_ALLOC_REGION] << endl;
#endif
    status=IC_CANT_ALLOC_REGION;
    return;
  }

  IC_Desc* desc=getPartiDescriptor(a); /*,thisProgramAO==otherProgramAO);*/
  if (!desc) {
#ifdef IC_DEBUG
    cout << IC_EndPoint::IC_errors[-IC_CANT_GET_DA_DESCRIPTOR] << endl;
#endif
    status=IC_CANT_GET_DA_DESCRIPTOR;
    return;
  }

  /* this can probably be cached in a more sophisticated version!!!*/
  IC_Sched* sched=IC_Compute_schedule(thisProgram,otherProgram,desc,region_set,1);
  if (!sched) {
#ifdef IC_DEBUG
    cout << IC_EndPoint::IC_errors[-IC_CANT_COMPUTE_COMM_SCHEDULE] << endl;
#endif
    status=IC_CANT_COMPUTE_COMM_SCHEDULE;
    return;
  }

  int ret=IC_Recv_$1(otherProgram,sched,a.getLocalArray().getDataPointer(), msgtag);
  if (ret!=0) {
#ifdef IC_DEBUG
    cout << IC_EndPoint::IC_errors[-IC_COMM_FAILURE] << endl;
#endif
    status=IC_COMM_FAILURE;
    return;
  }

  IC_Free_region(region_set[0]);
  IC_Free_desc(desc);

  status=IC_OK;
}'
)


define(`PPP_BCAST_LOCAL_HELPER',`
void IC_EndPoint::bcastLocalArray(const $1Array& a, int nelems, int tag, int& status) {
  
  status = IC_BCast_Local_$1(otherProgram, a.getLocalArray().getDataPointer(), nelems, tag); 

  if(status != 1){
#ifdef IC_DEBUG
    cout << IC_EndPoint::IC_errors[-IC_COMM_FAILURE] << endl;
#endif
    status=IC_COMM_FAILURE;
    return;
  }

  status = IC_OK;
}'
)

define(`PPP_RECV_LOCAL_HELPER',`
void IC_EndPoint::recvLocalArray(const $1Array& a, int nelems, int tag, int& status) {
 
  status = IC_Recv_Local_$1(otherProgram, a.getLocalArray().getDataPointer(), nelems, tag); 

  if(status != 1){
#ifdef IC_DEBUG
    cout << IC_EndPoint::IC_errors[-IC_COMM_FAILURE] << endl;
#endif
    status=IC_COMM_FAILURE;
    return;
  }

  status = IC_OK;
}'
)


define(`PPP_ICXJD_BCAST_LOCAL_HELPER',`
void IC_EndPointSet::bcastLocalArray(const char* ptname, const $1Array& arr, int nelems, int& stat){ 
    #define DEBUG

    bool found = false;

    map<string, IC_EndPoint*>::const_iterator p2ep_iter;
    map<string, int>::const_iterator p2tag_iter;

    IC_EndPoint* ep = NULL;
    IC_Program* otherprog = NULL;
    int tag = -1;

    p2ep_iter = this->port2ep.find(ptname);
    if(p2ep_iter != this->port2ep.end()){
        ep = (IC_EndPoint*)p2ep_iter->second;
        otherprog = ep->get_otherprog();
    }
    else{
        #ifdef IC_DEBUG
            cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_COMM_FAILURE] << endl;
        #endif
        stat = IC_EndPoint::IC_COMM_FAILURE;
        return;
    }

    p2tag_iter = this->port2tag.find(ptname);
    if(p2tag_iter != this->port2tag.end()){
        tag = (int)p2tag_iter->second; 
        #ifdef DEBUG
            cout << "port tag found : " << tag << endl;
        #endif
    }
    else{
        #ifdef IC_DEBUG
            cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_COMM_FAILURE] << endl;
        #endif
        stat = IC_EndPoint::IC_COMM_FAILURE;
        return;
    }


    stat = IC_BCast_Local_$1(otherprog, arr.getLocalArray().getDataPointer(), nelems, tag);
    if (stat != 1) {
        #ifdef IC_DEBUG
            cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_COMM_FAILURE] << endl;
        #endif
        stat = IC_EndPoint::IC_COMM_FAILURE;
        return;
    }

    stat = IC_OK;
}'
)


define(`PPP_ICXJD_RECV_LOCAL_HELPER',`
void IC_EndPointSet::recvLocalArray(const char* ptname, const $1Array& arr, int nelems, int& stat){ 
    #define DEBUG

    bool found = false;

    map<string, IC_EndPoint*>::const_iterator p2ep_iter;
    map<string, int>::const_iterator p2tag_iter;

    IC_EndPoint* ep = NULL;
    IC_Program* otherprog = NULL;
    int tag = -1;

    p2ep_iter = this->port2ep.find(ptname);
    if(p2ep_iter != this->port2ep.end()){
        ep = (IC_EndPoint*)p2ep_iter->second;
        otherprog = ep->get_otherprog();
    }
    else{
        #ifdef IC_DEBUG
            cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_COMM_FAILURE] << endl;
        #endif
        stat = IC_EndPoint::IC_COMM_FAILURE;
        return;
    }

    p2tag_iter = this->port2tag.find(ptname);
    if(p2tag_iter != this->port2tag.end()){
        tag = (int)p2tag_iter->second; 
        #ifdef DEBUG
            cout << "port tag found : " << tag << endl;
        #endif
    }
    else{
        #ifdef IC_DEBUG
            cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_COMM_FAILURE] << endl;
        #endif
        stat = IC_EndPoint::IC_COMM_FAILURE;
        return;
    }

    stat = IC_Recv_Local_$1(otherprog, arr.getLocalArray().getDataPointer(), nelems, tag);
    if (stat != 1) {
        #ifdef IC_DEBUG
            cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_COMM_FAILURE] << endl;
        #endif
        stat = IC_EndPoint::IC_COMM_FAILURE;
        return;
    }

    stat = IC_OK;
}'
)


define(`PPP_ARRAY_REGISTER_HELPER',`
void IC_EndPointSet::registerArray(const char* ptname, const $1Array& arr, int& stat){        
    string portname = ptname;
    stat = IC_ARRAY_REGISTER_FAILURE;        

    map<string, const $1Array*>::const_iterator iter;
    iter = this->ppp_$1_arrmap.find(portname);

    if(iter == this->ppp_$1_arrmap.end()){
        #ifdef DEBUG
            cout << "no such key in ppp_$1_arrmap..." << endl;
        #endif
        stat = IC_ARRAY_REGISTER_FAILURE;
        return;
    }
    else{
        /* bind given array into the array map        */
        #ifdef DEBUG
            cout << "portname found in ppp_$1_arrmap..." << endl;
        #endif
        this->ppp_$1_arrmap[portname] = &arr;
        stat = IC_OK;
    }

    if(stat == IC_OK){
        unsigned ndim=arr.numberOfDimensions();

        if (ndim > IC_MAX_DIM) {
            #ifdef IC_DEBUG
                cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_INVALID_NDIM] << endl;
            #endif
            stat = IC_EndPoint::IC_INVALID_NDIM;
            return;
        }

        unsigned i;
        int lower[IC_MAX_DIM], upper[IC_MAX_DIM], stride[IC_MAX_DIM];
  
        for(i=0;i<ndim;++i) {
            lower[i]=arr.getBase(i); 
            upper[i]=arr.getBound(i); 
            stride[i]=arr.getStride(i);
            
            #ifdef DEBUG
                cout << lower[i] << " " << upper[i] << " " << stride[i] << endl;
            #endif
        }

        #ifdef DEBUG
            cout << "number of dimensions " << ndim << endl;
        #endif

        IC_Region** region_set = (IC_Region**)malloc(sizeof(IC_Region*));
        region_set[0] = IC_Create_block_region(ndim, lower, upper, stride);
        if (!region_set[0]) {
            #ifdef IC_DEBUG
                cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_CANT_ALLOC_REGION] << endl;
            #endif
            
            stat = IC_EndPoint::IC_CANT_ALLOC_REGION;
            return;
        }
        else{
	        #ifdef DEBUG
	            cout << "register the regionset to port2regionset map" << endl;
	        #endif
        	this->port2regionset[portname] = region_set;
        }

        map<string, IC_EndPoint*>::const_iterator epiter;
        epiter = this->port2ep.find(portname);
        
        if(epiter == this->port2ep.end()){
            #ifdef DEBUG
                cout << "no such key in port2ep..." << endl;
            #endif
            
            stat = IC_ARRAY_REGISTER_FAILURE;
            return;
        }
        else{
            IC_EndPoint* epptr = (IC_EndPoint*)epiter->second;
            unsigned myao = epptr->get_myprog_arrayorder();
            unsigned oao  = epptr->get_otherprog_arrayorder();
            #ifdef DEBUG
                cout << "Array Order : " << myao << " " << oao <<endl;
            #endif
            
            IC_Desc* desc = getPartiDescriptor(arr); /*, myao==oao);*/
            
            if (!desc) {
                #ifdef IC_DEBUG
                    cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_CANT_GET_DA_DESCRIPTOR] << endl;
                #endif
                stat = IC_EndPoint::IC_CANT_GET_DA_DESCRIPTOR;
                return;
            }
            else{
		        #ifdef DEBUG
		            cout << "register the desc to port2desc" << endl;
		        #endif
                this->port2desc[portname] = desc;
            }
        }        
    }
    else{
        stat = IC_ARRAY_REGISTER_FAILURE;        
    }
}'
)


/*//// M4 macros for Fortran 90 methods*/

define(`F90_EXPORT_HELPER',`
void IC_EndPoint::exportArray(const F90_$1Array& a, int msgtag, int& status) {

/* exportArrays a double array using InterComm*/
  unsigned i;
  unsigned ndim=a.rank;
  if (ndim>IC_MAX_DIM) {
#ifdef IC_DEBUG
    cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_INVALID_NDIM] << endl;
#endif
    status=IC_EndPoint::IC_INVALID_NDIM;
    return;
  }

  int lower[IC_MAX_DIM], upper[IC_MAX_DIM], stride[IC_MAX_DIM];

  for(i=0;i<ndim;++i) {
    lower[i]=0;
    upper[i]= a.shape[i] - 1;
    stride[i]= 1;

#ifdef DEBUG    
    printf("Export: F90Array Structure: shape [%d] %d stride [%d] %d\n",i,a.shape[i],i,a.stride[i]);
#endif
  }

#ifdef DEBUG    
  cout << "number of dimensions is " << ndim << endl;
#endif


  IC_Region* region_set[1];
  region_set[0]=IC_Create_block_region(ndim,lower,upper,stride);
  if (!region_set[0]) {
#ifdef IC_DEBUG
    cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_CANT_ALLOC_REGION] << endl;
#endif
    status=IC_EndPoint::IC_CANT_ALLOC_REGION;
    return;
  }

  IC_Desc* desc=buildDescriptor(a);
  if (!desc) {
#ifdef IC_DEBUG
    cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_CANT_GET_DA_DESCRIPTOR] << endl;
#endif
    status=IC_EndPoint::IC_CANT_GET_DA_DESCRIPTOR;
    return;
  }

  /* this can probably be cached in a more sophisticated version!!!*/
  IC_Sched* sched=IC_Compute_schedule(thisProgram,otherProgram,desc,region_set,1);
  if (!sched) {
#ifdef IC_DEBUG
    cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_CANT_COMPUTE_COMM_SCHEDULE] << endl;
#endif
    status=IC_EndPoint::IC_CANT_COMPUTE_COMM_SCHEDULE;
    return;
  }

  int ret=IC_Send_$1(otherProgram,sched,a.base, msgtag);
  if (ret!=0) {
#ifdef IC_DEBUG
    cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_COMM_FAILURE] << endl;
#endif
    status=IC_EndPoint::IC_COMM_FAILURE;
    return;
  }

  IC_Free_sched(sched);
  IC_Free_region(region_set[0]);
  IC_Free_desc(desc);

  status=IC_OK;

}')

define(`F90_IMPORT_HELPER',`
void IC_EndPoint::importArray(const F90_$1Array& a, int msgtag, int& status) {

/* importArrays a double array using InterComm*/
  unsigned ndim=a.rank;
  if (ndim>IC_MAX_DIM) {
#ifdef IC_DEBUG
    cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_INVALID_NDIM] << endl;
#endif
    status=IC_EndPoint::IC_INVALID_NDIM;
    return;
  }
  unsigned i;
  int lower[IC_MAX_DIM], upper[IC_MAX_DIM], stride[IC_MAX_DIM];

  for(i=0;i<ndim;++i) {
    lower[i]=0;
    upper[i]= a.shape[i] - 1;
    stride[i]= 1;

#ifdef DEBUG
    printf("Import: F90Array Structure: shape [%d] %d stride [%d] %d\n",i, a.shape[i],i,a.stride[i]);
#endif
  }

#ifdef DEBUG
  cout << "number of dimensions is " << ndim << endl;
#endif

  IC_Region* region_set[1];
  region_set[0]=IC_Create_block_region(ndim,lower,upper,stride);
  if (!region_set[0]) {
#ifdef IC_DEBUG
    cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_CANT_ALLOC_REGION] << endl;
#endif
    status=IC_EndPoint::IC_CANT_ALLOC_REGION;
    return;
  }

  IC_Desc* desc=buildDescriptor(a);
  if (!desc) {
#ifdef IC_DEBUG
    cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_CANT_GET_DA_DESCRIPTOR] << endl;
#endif
    status=IC_EndPoint::IC_CANT_GET_DA_DESCRIPTOR;
    return;
  }

  /* this can probably be cached in a more sophisticated version!!!*/
  IC_Sched* sched=IC_Compute_schedule(thisProgram,otherProgram,desc,region_set,1);
  if (!sched) {
#ifdef IC_DEBUG
    cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_CANT_COMPUTE_COMM_SCHEDULE] << endl;
#endif
    status=IC_EndPoint::IC_CANT_COMPUTE_COMM_SCHEDULE;
    return;
  }

  int ret=IC_Recv_$1(otherProgram,sched,a.base, msgtag);
  if (ret!=0) {
#ifdef IC_DEBUG
    cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_COMM_FAILURE] << endl;
#endif
    status=IC_EndPoint::IC_COMM_FAILURE;
    return;
  }
 
  IC_Free_sched(sched);
  IC_Free_region(region_set[0]);
  IC_Free_desc(desc);

  status=IC_OK;

}')


define(`F90_BCAST_LOCAL_HELPER',`
void IC_EndPoint::bcastLocalArray(const F90_$1Array& a, int nelems, int tag, int& status) {
  
  status = IC_BCast_Local_$1(otherProgram, a.base, nelems, tag); 

  if(status != 0){
#ifdef IC_DEBUG
    cout << IC_EndPoint::IC_errors[-IC_COMM_FAILURE] << endl;
#endif
    status=IC_COMM_FAILURE;
    return;
  }

  status = IC_OK;
}'
)

define(`F90_RECV_LOCAL_HELPER',`
void IC_EndPoint::recvLocalArray(const F90_$1Array& a, int nelems, int tag, int& status) {
  
  status = IC_Recv_Local_$1(otherProgram, a.base, nelems, tag); 

  if(status != 0){
#ifdef IC_DEBUG
    cout << IC_EndPoint::IC_errors[-IC_COMM_FAILURE] << endl;
#endif
    status=IC_COMM_FAILURE;
    return;
  }

  status = IC_OK;
}'
)


define(`F90_ICXJD_BCAST_LOCAL_HELPER',`
void IC_EndPointSet::bcastLocalArray(const char* ptname, const F90_$1Array& arr, int nelems, int& stat){ 
    #define DEBUG

    bool found = false;

    map<string, IC_EndPoint*>::const_iterator p2ep_iter;
    map<string, int>::const_iterator p2tag_iter;

    IC_EndPoint* ep = NULL;
    IC_Program* otherprog = NULL;
    int tag = -1;

    p2ep_iter = this->port2ep.find(ptname);
    if(p2ep_iter != this->port2ep.end()){
        ep = (IC_EndPoint*)p2ep_iter->second;
        otherprog = ep->get_otherprog();
    }
    else{
        #ifdef IC_DEBUG
            cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_COMM_FAILURE] << endl;
        #endif
        stat = IC_EndPoint::IC_COMM_FAILURE;
        return;
    }

    p2tag_iter = this->port2tag.find(ptname);
    if(p2tag_iter != this->port2tag.end()){
        tag = (int)p2tag_iter->second; 
        #ifdef DEBUG
            cout << "port tag found : " << tag << endl;
        #endif
    }
    else{
        #ifdef IC_DEBUG
            cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_COMM_FAILURE] << endl;
        #endif
        stat = IC_EndPoint::IC_COMM_FAILURE;
        return;
    }

    stat = IC_BCast_Local_$1(otherprog, arr.base, nelems, tag);
    if (stat != 1) {
        #ifdef IC_DEBUG
            cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_COMM_FAILURE] << endl;
        #endif
        stat = IC_EndPoint::IC_COMM_FAILURE;
        return;
    } 

    stat = IC_OK;
}'
)

define(`F90_ICXJD_RECV_LOCAL_HELPER',`
void IC_EndPointSet::recvLocalArray(const char* ptname, const F90_$1Array& arr, int nelems, int& stat){ 
    #define DEBUG

    bool found = false;

    map<string, IC_EndPoint*>::const_iterator p2ep_iter;
    map<string, int>::const_iterator p2tag_iter;

    IC_EndPoint* ep = NULL;
    IC_Program* otherprog = NULL;
    int tag = -1;

    p2ep_iter = this->port2ep.find(ptname);
    if(p2ep_iter != this->port2ep.end()){
        ep = (IC_EndPoint*)p2ep_iter->second;
        otherprog = ep->get_otherprog();
    }
    else{
        #ifdef IC_DEBUG
            cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_COMM_FAILURE] << endl;
        #endif
        stat = IC_EndPoint::IC_COMM_FAILURE;
        return;
    }

    p2tag_iter = this->port2tag.find(ptname);
    if(p2tag_iter != this->port2tag.end()){
        tag = (int)p2tag_iter->second; 
        #ifdef DEBUG
            cout << "port tag found : " << tag << endl;
        #endif
    }
    else{
        #ifdef IC_DEBUG
            cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_COMM_FAILURE] << endl;
        #endif
        stat = IC_EndPoint::IC_COMM_FAILURE;
        return;
    }

    stat = IC_Recv_Local_$1(otherprog, arr.base, nelems, tag);
    if (stat != 1) {
        #ifdef IC_DEBUG
            cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_COMM_FAILURE] << endl;
        #endif
        stat = IC_EndPoint::IC_COMM_FAILURE;
        return;
    } 

    stat = IC_OK;
}'
)

define(`F90_BUILD_DESCRIPTOR_HELPER',`
IC_Desc* IC_EndPoint::buildDescriptor(const F90_$1Array& a) {
  if (static_cast<unsigned>(a.rank)>IC_MAX_DIM) {
#ifdef IC_DEBUG
    cout << IC_EndPoint::IC_errors[-IC_INVALID_NDIM] << endl;
#endif
    return NULL;
  }

  int tasks[1];
  tasks[0]=0;

  IC_Desc* desc;

  switch(a.rank) {
    case 1: {
      int blocks[1][2][1];
      int blockstride[1];

      blocks[0][0][0]=0;
      blocks[0][1][0]=a.shape[0]-1;

      blockstride[0] = a.stride[0];

      desc=create_f90_bdecomp_desc(a.rank,&blocks[0][0][0],tasks,1, blockstride, IC_EndPoint::IC_COLUMN_MAJOR);
      break;
    }
    case 2: {
      int blocks[1][2][2];
      int blockstride[2];

      blocks[0][0][0]=0;
      blocks[0][0][1]=0;

      blocks[0][1][0]=a.shape[0]-1;
      blocks[0][1][1]=a.shape[1]-1;

      blockstride[0] = a.stride[0];
      blockstride[1] = a.stride[1];

      desc=create_f90_bdecomp_desc(a.rank,&blocks[0][0][0],tasks,1, blockstride, IC_EndPoint::IC_COLUMN_MAJOR);

      break;
    }
    case 3: {
      int blocks[1][2][3];
      int blockstride[3];

      blocks[0][0][0]=0;
      blocks[0][0][1]=0;
      blocks[0][0][2]=0;

      blocks[0][1][0]=a.shape[0]-1;
      blocks[0][1][1]=a.shape[1]-1;
      blocks[0][1][2]=a.shape[2]-1;

      blockstride[0] = a.stride[0];
      blockstride[1] = a.stride[1];
      blockstride[2] = a.stride[2];

      desc=create_f90_bdecomp_desc(a.rank,&blocks[0][0][0],tasks,1, blockstride, IC_EndPoint::IC_COLUMN_MAJOR);

      break;
    }
    case 4: {
      int blocks[1][2][4];
      int blockstride[4];

      blocks[0][0][0]=0;
      blocks[0][0][1]=0;
      blocks[0][0][2]=0;
      blocks[0][0][3]=0;

      blocks[0][1][0]=a.shape[0]-1;
      blocks[0][1][1]=a.shape[1]-1;
      blocks[0][1][2]=a.shape[2]-1;
      blocks[0][1][3]=a.shape[3]-1;

      blockstride[0] = a.stride[0];
      blockstride[1] = a.stride[1];
      blockstride[2] = a.stride[2];
      blockstride[3] = a.stride[3];

      desc=create_f90_bdecomp_desc(a.rank,&blocks[0][0][0],tasks,1, blockstride, IC_EndPoint::IC_COLUMN_MAJOR);

      break;
    }

    default:
      assert(0);
  }

  return desc;
}')


define(`F90_ARRAY_REGISTER_HELPER',`
#define DEBUG
void IC_EndPointSet::registerArray(const char* ptname, const F90_$1Array& arr, int& stat){
    string portname = ptname;
    stat = IC_ARRAY_REGISTER_FAILURE;        

    map<string, const F90_$1Array*>::const_iterator iter;
    iter = this->f90_$1_arrmap.find(portname);

    if(iter == this->f90_$1_arrmap.end()){
        #ifdef DEBUG
            cout << "no such key in f90_$1_arrmap..." << endl;
        #endif
        stat = IC_ARRAY_REGISTER_FAILURE;
        return;
    }
    else{
        /* bind given array into the array map        */
        #ifdef DEBUG
            cout << "portname found in f90_$1_arrmap..." << endl;
        #endif
        this->f90_$1_arrmap[portname] = &arr;
        stat = IC_OK;
    }

    if(stat == IC_OK){
        unsigned ndim = arr.rank;
        
        if (ndim > IC_MAX_DIM) {
            #ifdef IC_DEBUG
                cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_INVALID_NDIM] << endl;
            #endif
            stat = IC_EndPoint::IC_INVALID_NDIM;
            return;
        }

        unsigned i;
        int lower[IC_MAX_DIM], upper[IC_MAX_DIM], stride[IC_MAX_DIM];
        
        for(i=0;i<ndim;++i) {
            lower[i] = 0; 
            upper[i] = arr.shape[i]-1; 
            stride[i] = 1;

            #ifdef DEBUG
                cout << "F90Array Structure: shape [" << i << "] " << arr.shape[i] << " stride[" << i << "] " << arr.stride[i] << endl;
            #endif
        }

        #ifdef DEBUG
          cout << "number of dimensions is " << ndim << endl;
        #endif

        /* create region*/
        IC_Region** region_set = (IC_Region**)malloc(sizeof(IC_Region*));
        region_set[0] = IC_Create_block_region(ndim, lower, upper, stride);
        if (!region_set[0]) {
            #ifdef IC_DEBUG
                cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_CANT_ALLOC_REGION] << endl;
            #endif

            stat = IC_EndPoint::IC_CANT_ALLOC_REGION;
            return;
        }
        else{
            #ifdef DEBUG
                cout << "register the regionset to port2regionset map" << endl;
            #endif
            this->port2regionset[portname] = region_set;
        }

        /* create desc */
        map<string, IC_EndPoint*>::const_iterator epiter;
        epiter = this->port2ep.find(portname);
        
        if(epiter == this->port2ep.end()){
            #ifdef DEBUG
                cout << "no such key in port2ep..." << endl;
            #endif
            
            stat = IC_ARRAY_REGISTER_FAILURE;
            return;
        }
        else{
            IC_EndPoint* epptr = (IC_EndPoint*)epiter->second;
            unsigned myao = epptr->get_myprog_arrayorder();
            unsigned oao  = epptr->get_otherprog_arrayorder();
            #ifdef DEBUG
                cout << "Array Order : " << myao << " " << oao <<endl;
            #endif
            
            IC_Desc* desc = IC_EndPoint::buildDescriptor(arr);
            
            if (!desc) {
                #ifdef IC_DEBUG
                    cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_CANT_GET_DA_DESCRIPTOR] << endl;
                #endif
                stat = IC_EndPoint::IC_CANT_GET_DA_DESCRIPTOR;
                return;
            }
            else{
                #ifdef DEBUG
                    cout << "register the desc to port2desc" << endl;
                #endif
                this->port2desc[portname] = desc;
            }
        }        
    }
    else{
        stat = IC_ARRAY_REGISTER_FAILURE;        
    }
}'
)


/*/// instantiatiated code*/
divert
#include "InterComm.h"
#include "Descriptor.h"
#include "IC_EndPoint.h"
#include "IC_EndPointSet.h"
#include <assert.h>
#include <iostream>

using namespace std;

#ifdef PPLUSPLUS
#include <A++.h>
#include "pplus2bparti.h"
#endif

#ifdef PPLUSPLUS

/* Methods for int arrays*/
PPP_EXPORT_HELPER(int)
PPP_IMPORT_HELPER(int)
PPP_BCAST_LOCAL_HELPER(int)
PPP_RECV_LOCAL_HELPER(int)
PPP_ICXJD_BCAST_LOCAL_HELPER(int)
PPP_ICXJD_RECV_LOCAL_HELPER(int)

/* Methods for float arrays*/
PPP_EXPORT_HELPER(float)
PPP_IMPORT_HELPER(float)
PPP_BCAST_LOCAL_HELPER(float)
PPP_RECV_LOCAL_HELPER(float)
PPP_ICXJD_BCAST_LOCAL_HELPER(float)
PPP_ICXJD_RECV_LOCAL_HELPER(float)

/* Methods for double arrays*/
PPP_EXPORT_HELPER(double)
PPP_IMPORT_HELPER(double)
PPP_BCAST_LOCAL_HELPER(double)
PPP_RECV_LOCAL_HELPER(double)
PPP_ICXJD_BCAST_LOCAL_HELPER(double)
PPP_ICXJD_RECV_LOCAL_HELPER(double)

/* Methods for arrays registration*/
PPP_ARRAY_REGISTER_HELPER(int)
PPP_ARRAY_REGISTER_HELPER(float)
PPP_ARRAY_REGISTER_HELPER(double)

#endif

#ifdef FORTRAN90

/* Methods for char arrays*/
F90_EXPORT_HELPER(char)
F90_IMPORT_HELPER(char)
F90_BCAST_LOCAL_HELPER(char)
F90_RECV_LOCAL_HELPER(char)
F90_ICXJD_BCAST_LOCAL_HELPER(char)
F90_ICXJD_RECV_LOCAL_HELPER(char)
F90_BUILD_DESCRIPTOR_HELPER(char)
F90_ARRAY_REGISTER_HELPER(char)

/* Methods for short arrays*/
F90_EXPORT_HELPER(short)
F90_IMPORT_HELPER(short)
F90_BCAST_LOCAL_HELPER(short)
F90_RECV_LOCAL_HELPER(short)
F90_ICXJD_BCAST_LOCAL_HELPER(short)
F90_ICXJD_RECV_LOCAL_HELPER(short)
F90_BUILD_DESCRIPTOR_HELPER(short)
F90_ARRAY_REGISTER_HELPER(short)

/* Methods for int arrays*/
F90_EXPORT_HELPER(int)
F90_IMPORT_HELPER(int)
F90_BCAST_LOCAL_HELPER(int)
F90_RECV_LOCAL_HELPER(int)
F90_ICXJD_BCAST_LOCAL_HELPER(int)
F90_ICXJD_RECV_LOCAL_HELPER(int)
F90_BUILD_DESCRIPTOR_HELPER(int)
F90_ARRAY_REGISTER_HELPER(int)

/* Methods for float arrays*/
F90_EXPORT_HELPER(float)
F90_IMPORT_HELPER(float)
F90_BCAST_LOCAL_HELPER(float)
F90_RECV_LOCAL_HELPER(float)
F90_ICXJD_BCAST_LOCAL_HELPER(float)
F90_ICXJD_RECV_LOCAL_HELPER(float)
F90_BUILD_DESCRIPTOR_HELPER(float)
F90_ARRAY_REGISTER_HELPER(float)

/* Methods for double arrays*/
F90_EXPORT_HELPER(double)
F90_IMPORT_HELPER(double)
F90_BCAST_LOCAL_HELPER(double)
F90_RECV_LOCAL_HELPER(double)
F90_ICXJD_BCAST_LOCAL_HELPER(double)
F90_ICXJD_RECV_LOCAL_HELPER(double)
F90_BUILD_DESCRIPTOR_HELPER(double)
F90_ARRAY_REGISTER_HELPER(double)

#endif
