/**
 * $Author: normanlo $
 * $Date: 2006/05/02 21:23:21 $
 * $Revision: 1.1.2.31 $
 * $Source: /fs/schisto7/parti/repository/InterComm/src/Attic/xjd_handle.h,v $
 * 
 * Written by Il-Chul Yoon <iyoon@cs.umd.edu>
 */
#ifndef XJDHANDLE_H
#define XJDHANDLE_H

/** following definitions are used for parsing XJD file
 * these must be same with the definitions at IC_EndPointSet.h
/* reserved define for ezxml */
#define XJDVERSION  "version"
#define COMPLIST "components"
#define COMP     "component"
#define CONNLIST "connections"
#define CONN     "connection"
/* reserved define for IC_XJDRegion */
#define EXPORTER "exporter"
#define IMPORTER "importer"
#define EXPPORT  "exportport"
#define IMPPORT  "importport"
#define DATATYPE "type"
#define TAG      "msgtag"
#define COMMTYPE "commtype"
/* reserved define for IC_XJDComponent */
#define COMPID       "id"
#define COMPNAME     "comp_id"
#define N_NODE       "nNode"

/**
 * reserved return codes from XJD lib 
 * NOT FIXED YET
 */
#define IC_OK                   0
#define IC_INITIALIZE_OK        0
#define XJDCOMP_REGISTER_OK     0
#define XJDREGION_REGISTER_OK   0
#define XJDREGION_COMMIT_OK     0
#define XJD_VALIDATE_OK         0

#define XJDCOMP_REGISTER_ERR    -1
#define XJDREGION_REGISTER_ERR  -1
#define IC_INITPVM_ERR          -1
#define IC_INITMPI_ERR          -1    /* non-used yet*/
#define IC_INITIALIZE_FAILED    -1   
#define IC_COMPUTE_SCHED_FAILED -1 
#define IC_IMPORT_FAILED        -1
#define IC_EXPORT_FAILED        -1
#define XJD_VALIDATE_ERR        -1


/**
 * IC_Initialize for XJD-based programming API
 */
void IC_Initialize_1_5(IC_XJD* xjd, char* pName, int rank, int* icsts);

/**
 * IC_Initialize for API using SML2
 */
void IC_Initialize_2_0();

/**
 * IC_Initialize for API using SML2
IC_RegionHandle* IC_InitHandle_1_5();
 */


/**
 * IC_Initialize for API using SML2
IC_RegionHandle* IC_InitHandle_2_0();
 */


/**
 * IC_Register_region for XJD-based programming API
 */
void IC_Register_region_1_5(IC_XJD* xjd, IC_Region** rgnset, int set_size, char* set_name, IC_Desc* desc, void* local_data, int* icsts);


/**
 * IC_Register_region for API using SML2 
 */
void IC_Register_region_2_0(IC_XJD* xjd, IC_Region** rgnset, int set_size, char* set_name, IC_Desc* desc, void* local_data, int* icsts);


/**
 * IC_Commit_region for XJD-based programming API
 */
void IC_Commit_region_1_5(IC_XJD* xjd, int* icsts);


/**
 * IC_Commit_region for API using SML2
 */
void IC_Commit_region_2_0();


/**
 * IC_Export for XJD-based programming API
 */
int IC_Export_1_5(IC_XJD* xjd, char* rgn_name);


/**
 * IC_Export for API using SML2
 */
int IC_Export_2_0(char* rgn_name);


/**
 * IC_Import for XJD-based programming API
 */
int IC_Import_1_5(IC_XJD* xjd, char* rgn_name);


/**
 * IC_Import for API using SML2
 */
int IC_Import_2_0(char* rgn_name);


/**
 * IC_BCast_Local for XJD-based programming API 
 * Maybe, can be used for SML2 based programming
 */
int IC_BCast_Local_1_5(IC_XJD* xjd, char* rgnset_name, void* data, int nelems);


/**
 * IC_Recv_Local for XJD-based programming API 
 * Maybe, can be used for SML2 based programming
 */
int IC_Recv_Local_1_5(IC_XJD* xjd, char* rgnset_name, void* data, int nelems);


/**
 * IC_Free_RegionHandle for XJD-based programming API
void IC_Free_RegionHandle_1_5(IC_RegionHandle* handle, int* icsts);
 */


/**
 * IC_Free_RegionHandle for API using SML2
void IC_Free_RegionHandle_2_0(IC_RegionHandle* handle, int* icsts);
 */


/**
 * IC_Finalize for XJD-based programming API
 */
void IC_Finalize_1_5(IC_XJD* xjd, int* icsts);


/**
 * IC_Finalize for API using SML2
 */
void IC_Finalize_2_0();


/**
 * Inspect connectoin information in IC_XJD and register region informations
 */
int inspect_connections(const char* pName, IC_XJD* xjd);


/**
 * Add a new component from IC_XJD into the component list
 * Assume that 'me' becomes the head of this list
 */
void add_comp(IC_XJDComponent* comp, IC_XJD* xjd);


/**
 * Add a new region into the region list
 */
void add_region(IC_XJDRegion* region, IC_XJD* xjd);


/**
 * Register program 'pName' into the IC_XJDComponent array of this 
 * program. The components related by region connections should 
 * be registered.
 */
int register_program(const char* pName, IC_XJD* xjd);


/**
 * Search a component with given name inside the component list
 * and return the related IC_Program 
 */
IC_Program* search_program(char* name, IC_XJD* xjd);


/**
 * Search a component name with given id inside the component list
 * and return the related program name 
 */
char* search_pname_byID(char* cid, IC_XJD* xjd);


/**
 * Search a region with given region name inside the region  list
 */IC_XJDRegion* search_region(char* name, IC_XJD* xjd);

/**
 * Register regions defined for each component based on XJD
 */
int register_region(
    char* myname, char* partnername,
    char* my_region_name, char* partner_region_name, char* type, char* commtype, int tag, IC_XJD* xjd);


/**
 * Deprecated?
 * Parse command-line options '-m' is for this program name
 * and '-d' is for XJD file name
 */
int parse_param(int argc, char **argv, char** xjd);


/**
 * Print all components in the global IC_XJDComponent list
 */
void print_components(IC_XJD* xjd);


/**
 * Print a component
 */
void print_comp(IC_XJDComponent* tmp );


/**
 * Print all regions in the global IC_XJDRegion list
 */
void print_regions(IC_XJD* xjd);


/**
 * Print region list
 */
void print_region(IC_XJDRegion* tmp);


/**
 * Return the first component in component definitions in XJD
 */
ezxml_t get_first_comp(ezxml_t xmlRoot);


/**
 * Release allocated memory for IC_XJDComponent recursively
 */
void release_comp(IC_XJDComponent* ptr, int first);


/**
 * Release allocated memory for IC_XJDRegion recursively
 */
void release_rgn(IC_XJDRegion* ptr);


/**
 * Return the XJD version info
 */
char* get_version(ezxml_t xmlRoot);

/**
 * Return the first component in component definitions in XJD
 */
ezxml_t get_first_comp(ezxml_t xmlRoot);


/**
 * Return the first connection in connection definitions in XJD
 */
ezxml_t get_first_conn(ezxml_t xmlRoot);


/**
 * Read XJD file name and return ezxml_t root object 
 */
ezxml_t read_xjd(const char* xjdname);

/**
 * Free the global XJD structure
 */
void free_xjd(ezxml_t xjd);

#endif
