#include <assert.h>
#include <errno.h>
#include <iostream>
#include <pthread.h>
#include <stdio.h>
#include <A++.h>
#include "testinter.h"

/* this is a simple example that can show how the Interpolator API can be used
   it can be used with two different endpoint abstractions: communication
   over files or communication using MetaChaos.

   ICENDPOINThas to be turned on in the Makefile*/

#if (!defined(ICENDPOINT))
#error You have to define ICENDPOINT
#endif

#ifdef ICENDPOINT
#include "IC_EndPointSet.h"
#endif

#ifdef PPLUSPLUS
#include "IC_util.h"
#include "IC_support.h"
Mutex MPIm; /* MPI is not reentrant!*/
#endif

Index I(3,3), J(2,3);
Index K(0,2), L(4,2);
Index M(0,3), N(0,3);
Index O(1,3), P(1,3);
const unsigned NITERATIONS=10;


/* this function simulates a parallel simulation code named simu1*/
void* simu1(void* xjdfilename) {
    int ic_err;
    int msgtag1 = 1001;
    int msgtag2 = 1002;
    int msgtag3 = 1003;

    #define SIMU1_NAME  "simu1"
    #define SIMU1_ASUB1 "Asub1"
    #define SIMU1_ASUB2 "Asub2"
    #define SIMU1_ASUB3 "Asub3"

    std::cout << ">>>> simu1 has started...\n";

    #ifdef PPLUSPLUS
        MPIm.lock();
    #endif
        doubleArray A(6,6);
        A=45;
        A.display("A");
    #ifdef PPLUSPLUS
        MPIm.unlock();
    #endif

    /**
     * initialize endpoint set
     */
    IC_EndPointSet epset((const char*)xjdfilename, (const char*)SIMU1_NAME, ic_err);
    assert(ic_err==IC_EndPointSet::IC_OK);
    
    /**
     * register arrays (bind physical data with array name)
     */
    const doubleArray& t1=A(I,J);
    epset.registerArray((const char*)SIMU1_ASUB1, t1, ic_err);
    assert(ic_err==IC_EndPointSet::IC_OK);
    
    const doubleArray& t2=A(M,N);
    epset.registerArray((const char*)SIMU1_ASUB2, t2, ic_err);
    assert(ic_err==IC_EndPointSet::IC_OK);

    /**
     * commit registered arrays
     */
    epset.commitArrays(ic_err);
    assert(ic_err==IC_EndPointSet::IC_OK);

    /**
     * transfer data
     */    
    epset.exportArray(SIMU1_ASUB1, ic_err);
    assert(ic_err==IC_EndPoint::IC_OK);
    std::cout << "A - first export is complete\n";

    epset.importArray(SIMU1_ASUB2, ic_err);
    assert(ic_err==IC_EndPoint::IC_OK);
    std::cout << "A - first import is complete\n";
    
    A.display("A");

    /**
     * broadcasting data
     */
    doubleArray broadA(5);
    for (int i=0; i<5; i++)
      broadA(i)=i*2;
    broadA.display("broadA");
    epset.bcastLocalArray(SIMU1_ASUB3, broadA, 5, ic_err);

    assert(ic_err==IC_EndPointSet::IC_OK);

    
  return NULL;
}

/* this function simulates a parallel simulation code named simu2*/
void* simu2(void* xjdfilename) {
    int ic_err;
    int msgtag1 = 1001;
    int msgtag2 = 1002;
    int msgtag3 = 1003;

    #define SIMU2_NAME  "simu2"
    #define SIMU2_BSUB1 "Bsub1"
    #define SIMU2_BSUB2 "Bsub2"
    #define SIMU2_BSUB3 "Bsub3"

    std::cout << ">>>> simu2 has started...\n";

    #ifdef PPLUSPLUS
        MPIm.lock();
    #endif
        doubleArray B(6,6);
        B=20;
        B.display("B");
    #ifdef PPLUSPLUS
        MPIm.unlock();
    #endif

    /**
     * initialize endpoint set
     */
    IC_EndPointSet epset((const char*)xjdfilename, (const char*)SIMU2_NAME, ic_err);
    assert(ic_err==IC_EndPointSet::IC_OK);

    /**
     * register arrays (bind physical data with array name)
     */
    const doubleArray& t1=B(M,N);
    epset.registerArray((const char*)SIMU2_BSUB1, t1, ic_err);
    assert(ic_err==IC_EndPointSet::IC_OK);

    const doubleArray& t2=B(O,P);
    epset.registerArray((const char*)SIMU2_BSUB2, t2, ic_err);
    assert(ic_err==IC_EndPointSet::IC_OK);

    /**
     * commit registered arrays
     */
    epset.commitArrays(ic_err);
    assert(ic_err==IC_EndPointSet::IC_OK);

    /**
     * transfer data
     */    
    /*    epset.importArray(SIMU2_BSUB1, msgtag1, ic_err);    */
    epset.importArray(SIMU2_BSUB1, ic_err);
    assert(ic_err==IC_EndPoint::IC_OK);
    std::cout << "B - first import is complete\n";

    /*    epset.exportArray(SIMU2_BSUB2, msgtag2, ic_err);*/
    epset.exportArray(SIMU2_BSUB2, ic_err);
    assert(ic_err==IC_EndPoint::IC_OK);
    std::cout << "B - first export is complete\n";

    B.display("B");

    /*
     * Receiving broadcasting calls
     */
    doubleArray broadB(5);
    epset.recvLocalArray(SIMU2_BSUB3, broadB, 5, ic_err);


    assert(ic_err==IC_EndPoint::IC_OK);

    int test = 1;
    for (int i=0; i<5; i++)
      if (broadB(i) != i*2)
	test = 0;
    
    if (test == 1)
      std::cout << "Receiving Broadcast successfully!" << endl;
    else
      std::cout << "Receiving Broadcast FAILED!" << endl;
  
  return NULL;
}

/* this function simulates a parallel simulation code named simu1*/
void* simu3(void* xjdfilename) {
    int ic_err;
    int msgtag1 = 1001;
    int msgtag2 = 1002;
    int msgtag3 = 1003;

    #define SIMU3_NAME  "simu3"
    #define SIMU3_CSUB1 "Csub1"
    #define SIMU3_CSUB2 "Csub2"

    std::cout << ">>>> simu3 has started...\n";

    #ifdef PPLUSPLUS
        MPIm.lock();
    #endif
        doubleArray C(6,6);
        C=90;
        C.display("C");
    #ifdef PPLUSPLUS
        MPIm.unlock();
    #endif

    /**
     * initialize endpoint set
     */
    IC_EndPointSet epset((const char*)xjdfilename, (const char*)SIMU3_NAME, ic_err);
    assert(ic_err==IC_EndPointSet::IC_OK);


    cout << "*****Before registering Array " <<endl;
    /**
     * register arrays (bind physical data with array name)
     */
    const doubleArray& t1=C(M,N);
    epset.registerArray((const char*)SIMU3_CSUB1, t1, ic_err);
    assert(ic_err==IC_EndPointSet::IC_OK);

    const doubleArray& t2=C(O,P);
    epset.registerArray((const char*)SIMU3_CSUB2, t2, ic_err);
    assert(ic_err==IC_EndPointSet::IC_OK);

    /**
     * commit registered arrays
     */

    cout << "******** Before committing Array " <<endl;
    epset.commitArrays(ic_err);
    assert(ic_err==IC_EndPointSet::IC_OK);

    /**
     * transfer data
     */    
    /*    epset.importArray(SIMU3_CSUB1, msgtag2, ic_err);*/
    epset.importArray(SIMU3_CSUB1, ic_err);
    assert(ic_err==IC_EndPoint::IC_OK);
    std::cout << "C - first import is complete\n";

    /*    epset.exportArray(SIMU3_CSUB2, msgtag3, ic_err);*/
    epset.exportArray(SIMU3_CSUB2, ic_err);
    assert(ic_err==IC_EndPoint::IC_OK);
    std::cout << "C - first export is complete\n";

    C.display("C");

  return NULL;
}


int main(int argc, char* argv[]) {
#ifdef ICENDPOINT
  if (argc==1) {
    std::cerr << "Can't run both simu2 and simu1 as threads...\n";
    std::cerr << "Try '" << argv[0] << " simu2' or '" << argv[0] << " simu1' << or '" << argv[0] << " simu3'!\n";
    return -1;
  }
#endif

  if (argc>1 && 
      (strcmp(argv[1],"simu2")!=0 && strcmp(argv[1],"simu1")!=0) && strcmp(argv[1],"simu3")!=0) {
    std::cerr << "Try '" << argv[0] << " simu2' or '" << argv[0] << " simu1' << or '" << argv[0] << " simu3'!\n";
    return -1;
  }


#ifdef PPLUSPLUS
  std::cout << ">> P++ is being used...\n";
#else
  std::cout << ">> A++ is being used...\n";
#endif

#ifdef ICENDPOINT
  std::cout << ">> IC_ENDPOINT is being used...\n";
#endif

  int iNumProcs=1;
  
  /* if first parameter is empty string, argv[0] is used as application name
     for A++/P++
     I did not find other use of argc and argv in InInitialize_Virtual_Machine*/
  Optimization_Manager::Initialize_Virtual_Machine("", iNumProcs, argc, argv);


  pthread_attr_t thattr;
  pthread_t tid1, tid2, tid3;
  int ret=pthread_attr_init(&thattr);
  assert(ret==0);
  ret=pthread_attr_setdetachstate(&thattr,PTHREAD_CREATE_JOINABLE);
  assert(ret==0);


  if (strcmp(argv[1],"simu1")==0) {
    ret=pthread_create(&tid1, &thattr, simu1, (void*)"./simulation.xjd");
    assert(ret==0);
  }

  if (strcmp(argv[1],"simu2")==0) {
    ret=pthread_create(&tid2, &thattr, simu2, (void*)"./simulation.xjd");
    assert(ret==0);
  }

  if (strcmp(argv[1],"simu3")==0) {
    ret=pthread_create(&tid3, &thattr, simu3, (void*)"./simulation.xjd");
    assert(ret==0);
  }

  if (argc==1 || strcmp(argv[1],"simu1")==0) {
    ret=pthread_join(tid1,NULL);
    if(ret!=0) {
      perror("waiting for simu1 to finish:");
      return -1;
    }
  }

  if (argc==1 || strcmp(argv[1],"simu2")==0) {
    ret=pthread_join(tid2,NULL);
    if(ret!=0) {
      perror("waiting for simu2 to finish:");
      return -1;
    }
  }

  if (argc==1 || strcmp(argv[1],"simu3")==0) {
    ret=pthread_join(tid3,NULL);
    if(ret!=0) {
      perror("waiting for simu3 to finish:");
      return -1;
    }
  }

  Optimization_Manager::Exit_Virtual_Machine();

  return 0;
}
